#include "eyedropper.h"

SwatchWindow::SwatchWindow()
	: BWindow( BRect( 0, 0, 100, 40 ),
	"Eyedropper",
	B_NO_BORDER_WINDOW_LOOK,
	B_FLOATING_ALL_WINDOW_FEEL,
	B_NOT_ZOOMABLE | B_ASYNCHRONOUS_CONTROLS | B_NOT_CLOSABLE | B_NOT_ANCHORED_ON_ACTIVATE | B_NO_WORKSPACE_ACTIVATION)
{	
	_ownerMessenger = NULL;
	_hexToClipboard = true;	
	_sv = NULL;
	_asleep = true;
	_clickToQuit = true;
}

SwatchWindow::~SwatchWindow()
{
	;
}

void SwatchWindow::SetOwner(BView *owner)
{
	_ownerMessenger = new BMessenger(owner);
}

void SwatchWindow::MessageReceived(BMessage* message)
{
	switch(message->what)
	{

		case Z_GO_AWAY:
		{
			GoToSleep();
			break;
		}
		
		case Z_WAKE_UP:
		{
			WakeUp();
			break;
		}
		
		case Z_SET_CLICK_TO_QUIT:
		{
			bool ctq;
			if (message->FindBool(EYEDROPPER_CLICK_TO_QUIT, &ctq) == B_OK)
				_clickToQuit = ctq;
			if (!IsHidden()) Activate();
			break;
		}
		
		
		case Z_SET_COPY_MODE:
		{
			bool mode;
			if (message->FindBool(EYEDROPPER_COPY_MODE, &mode) == B_OK)
				_hexToClipboard = mode;

			break;
		}

	
		default:
		{
			BWindow::MessageReceived(message);
			break;
		}
	}
}

void SwatchWindow::WindowActivated(bool active)
{
	if (_clickToQuit && !active)
	{
		if (!_asleep)
		{
			//write color to clipboard
			BMessage *clip;
			if (be_clipboard->Lock())
			{
				be_clipboard->Clear();
				clip = be_clipboard->Data();
				if (clip)
				{
					BString color = (_hexToClipboard ? _sv->GetHEXString() : _sv->GetRGBString() );
					clip->AddData("text/plain", B_MIME_TYPE, color.String(), color.Length());
					be_clipboard->Commit();
				}
				be_clipboard->Unlock();
			}

			//then hide
			GoToSleep();
		}
	}
}

void SwatchWindow::WakeUp()
{
	if (!_asleep) return;
	
	_asleep = false;

	if (_sv == NULL)
	{		
		_sv = new SwatchView(Bounds());
		AddChild(_sv);	
	}
	else
	{
		AddChild(_sv);	
	}

	Show();
}

void SwatchWindow::GoToSleep()
{
	_asleep = true;
	if (_sv) RemoveChild(_sv);
	if (!IsHidden()) Hide();
	
	if (_ownerMessenger) _ownerMessenger->SendMessage(EYEDROPPER_WENT_TO_SLEEP);
}


//*******************************

SwatchView::SwatchView(BRect frame)
	: BView(frame, "SwatchView", B_FOLLOW_ALL, B_WILL_DRAW | B_FRAME_EVENTS)
	{
	
	_screenRect.Set(0,0, 4, 4);
	_screenSnip = NULL;	
	SetEventMask(B_POINTER_EVENTS, B_NO_POINTER_HISTORY);	
}

SwatchView::~SwatchView()
{

}

void SwatchView::Draw(BRect updateRect )
{	
	BRect bnds(Bounds());
	SetDrawingMode(B_OP_BLEND);
	SetHighColor(0,0,0,255);
	StrokeRect(bnds);
	SetHighColor(255, 255, 255, 255);
	StrokeLine(bnds.LeftBottom(), bnds.RightBottom());
	StrokeLine(bnds.RightTop(), bnds.RightBottom());	
	
	float w = StringWidth(&_hexColorString[2]); //skip first two chars, since sprintf writes four hex doubles for an int32
	font_height h;
	GetFontHeight(&h);
	
	BPoint p(Bounds().Width() / 2.0 - w/2.0, Bounds().Height() / 4.0 + (h.ascent + h.descent)/4.0);

	SetDrawingMode(B_OP_INVERT);
	DrawString(&_hexColorString[2], p);	

	w = StringWidth(_rgbColorString);
	p.Set(Bounds().Width() / 2.0 - w/2.0, Bounds().Height() * 3.0 / 4.0 + (h.ascent + h.descent)/4.0);
	DrawString(_rgbColorString, p);
}

void SwatchView::AttachedToWindow()
{
	//force it to RGB32 to simplify math, because I'm dumb/lazy
	_screenSnip = new BBitmap(_screenRect, B_RGB32, false, true);
	
	BFont theFont(be_fixed_font);
	
	theFont.SetSize(12.0);
	theFont.SetFace(B_BOLD_FACE);
	SetFont(&theFont, B_FONT_SIZE | B_FONT_FAMILY_AND_STYLE);
}

void SwatchView::MessageReceived(BMessage* message)
{
	switch (message->what)
	{
	
	
		default:
		{
			BView::MessageReceived(message);
			break;
		}
	}

}

void SwatchView::MouseMoved(BPoint p, uint32 transit, const BMessage *msg)
{
	ConvertToScreen(&p);
	BScreen screen;
	BRect screenFrame = screen.Frame();
	BRect swatchBounds = Bounds();
	float dx = 17, dy = 17;

	float h = _screenRect.Height(), w = _screenRect.Width();
	float left, top, right, bottom;
	
	//position the screenRect with mouse at upper left
	//but don't let screenRect move outside of screen
	
	left = p.x;
	right = left + w;
	
	top = p.y;
	bottom = top + h;
	
	//make certain _screenRect does not go offscreen
	if (right > screenFrame.right)
	{
		right = screenFrame.right;
		left = right - w;
	}
	
	if (bottom > screenFrame.bottom)
	{
		bottom = screenFrame.bottom;
		top = bottom - h;
	}	

	_screenRect.Set(left, top, right, bottom);

	//now grab the snapshot from the screen
	screen.ReadBitmap(_screenSnip, false, &_screenRect);
	
	//we want to position the window, but not to move it offscreen
	if ((p.x + dx + swatchBounds.Width()) > screenFrame.Width())
	{
		dx = - swatchBounds.Width() - dx;
	}

	if ((p.y + dy + swatchBounds.Height()) > screenFrame.Height())
	{
		dy = - swatchBounds.Height() - dy;
	}	
	Window()->MoveTo(BPoint(p.x + dx, p.y + dy));
	
	//the bitmap is set, default, to rgb32, so we don't need to
	//verify the color_space
	uint32 *bits = (uint32 *) _screenSnip->Bits();
	uint32 pixel = *(bits); //grab the first pixel (upper left corner)
	
	_dropColor.red = (pixel & 0x00FF0000) >> 16;
	_dropColor.green = (pixel & 0x0000FF00) >> 8;
	_dropColor.blue = (pixel & 0x000000FF);
	
	//write our color into the rgb and hex strings
	sprintf(_rgbColorString, "%d %d %d", _dropColor.red, _dropColor.green, _dropColor.blue);
	sprintf(_hexColorString, "%.8lx", pixel);
	
	SetViewColor(_dropColor);
	Invalidate();
}

BString SwatchView::GetRGBString()
{
	return BString(_rgbColorString);
}

BString SwatchView::GetHEXString()
{
	return BString(&_hexColorString[2]);
}