#include "magnifier.h"
#include <WindowScreen.h>

#define _ABS(a) (((a) > 0) ? (a) : -(a))

MagnifyWindow::MagnifyWindow()
	: BWindow( BRect( 0, 0, 100, 100 ),
	"Magnifier",
	B_NO_BORDER_WINDOW_LOOK,
	B_FLOATING_ALL_WINDOW_FEEL,
	B_NOT_ZOOMABLE | B_ASYNCHRONOUS_CONTROLS | B_NOT_CLOSABLE | B_NOT_ANCHORED_ON_ACTIVATE | B_NO_WORKSPACE_ACTIVATION)
{
	_ownerMessenger = NULL;
	_clickToQuit = false;
	_mv = NULL;
	_asleep = true;
}

MagnifyWindow::~MagnifyWindow()
{
}

void MagnifyWindow::SetOwner(BView *owner)
{
	_ownerMessenger = new BMessenger(owner);
}

void MagnifyWindow::MessageReceived(BMessage* message)
{
	switch(message->what)
	{
		case Z_GO_AWAY:
		{
			GoToSleep();		
			break;
		}
		
		case Z_WAKE_UP:
		{
			WakeUp();
			break;
		}
		
		case Z_SET_CLICK_TO_QUIT:
		{
			//also fired when settings windo wis created
			bool ctq;
			if (message->FindBool(MAGNIFY_CLICK_TO_QUIT, &ctq) == B_OK)
				_clickToQuit = ctq;
				
			if (!IsHidden()) Activate();
			break;
		}
	
		case Z_SET_MAGNIFICATION:
		{
			//likely to be called by settings window as we change zoom levels
			//also fired when settings windo wis created
			if (message->FindDouble(MAGNIFY_MAG_LEVEL, &_magnification) == B_OK)
			{
				if (_mv) _mv->SetMagnification(_magnification);
			}
			
			break;
		}
		
		case Z_SET_WINDOW_SIZE:
		{
			//called by settings window as size slider is dragged -- BeOS! Live updating baby!
			//also fired off when settings window is created
			BRect frame;
			if (message->FindRect(MAGNIFY_WINDOW_SIZE, &frame) == B_OK)
			{
				ResizeTo(frame.Width(), frame.Height());
				//magnification needs to be updated when frame is resized
				if (_mv) _mv->SetMagnification(_magnification);
			}
			break;
		}
	
	
		default:
		{
			BWindow::MessageReceived(message);
			break;
		}
	}
}

void MagnifyWindow::WindowActivated(bool active)
{
	if (_clickToQuit && !active)
		GoToSleep();
}

void MagnifyWindow::WakeUp()
{
	if (!_asleep) return;
	
	_asleep = false;
	if (_mv == NULL)
	{	
		_mv = new MagnifyView(Bounds());
		AddChild(_mv);
		_mv->SetMagnification(_magnification);
		_mv->StartWatching();
	}
	else
	{
		_mv->StartWatching();
	}
	
	Show();
}

void MagnifyWindow::GoToSleep()
{
	_asleep = true;
	if (_mv) _mv->StopWatching();	
	if (!IsHidden()) Hide();	
	
	if (_ownerMessenger)_ownerMessenger->SendMessage(MAGNIFY_WENT_TO_SLEEP);
}


//*******************************


MagnifyView::MagnifyView(BRect frame)
	: BView(frame, "MagnifyView", B_FOLLOW_ALL, B_WILL_DRAW | B_FRAME_EVENTS)
	{

	_screenSnip = NULL;	
	
	SetMagnification(5);

	SetEventMask(B_POINTER_EVENTS | B_KEYBOARD_EVENTS, B_NO_POINTER_HISTORY);
	SetViewColor(B_TRANSPARENT_COLOR);

	_mr = NULL;	
	_shiftDown = false;
}

MagnifyView::~MagnifyView()
{
	if (_mr) delete _mr; //kill the message runner
	if (_screenSnip) delete _screenSnip;
}

void MagnifyView::AttachedToWindow()
{	
	//we want this message runner to run as long as this object exists
	//probably want to make it stop when MagnifyWindow goes to sleep...
	SetDrawingMode(B_OP_COPY);
	StartWatching();
}

void MagnifyView::StartWatching()
{
	if (_mr) return; //don't do anything if message runner already exists
	_mr = new BMessageRunner(this, new BMessage(Z_UPDATE_SNIPPET), 100000, -1);	
}

void MagnifyView::StopWatching()
{
	if (_mr)
	{
		delete _mr;
		_mr = NULL;
	}
}

void MagnifyView::MessageReceived(BMessage* message)
{
	switch (message->what)
	{
		case Z_UPDATE_SNIPPET:
		{
			//handler for message runner we've created
			//just pretend mouse has moved, but pass in the mouse's current location
			MouseMoved(_mouseLoc, 0, NULL);
			break;
		}
			
		default:
		{
			BView::MessageReceived(message);
			break;
		}
	}

}

void MagnifyView::Draw(BRect r)
{
	BRect bnds(Bounds());
		
	BRegion inner(bnds.InsetByCopy(1, 1));
	BRegion all(bnds);

	//to avoid annoying flashing for the instant the bitmap is drawn over the border	
	ConstrainClippingRegion(&inner);
	DrawBitmap(_screenSnip, bnds);
	ConstrainClippingRegion(&all);

	//draw inset frame
	SetHighColor(64,64,64,255);
	StrokeLine(bnds.LeftBottom(), bnds.LeftTop());
	StrokeLine(bnds.LeftTop(), bnds.RightTop());	
	
	SetHighColor(200, 200, 200, 255);
	StrokeLine(bnds.LeftBottom(), bnds.RightBottom());
	StrokeLine(bnds.RightTop(), bnds.RightBottom());	
}

void MagnifyView::MouseMoved(BPoint p, uint32 transit, const BMessage *msg)
{
	//we remeber this point unconverted
	BPoint cleanPoint = p;
	
	BScreen screen;
	BRect screenFrame = screen.Frame();
	BRect swatchBounds = Bounds();
		
	ConvertToScreen(&p);

	_checkForShiftKey(&p);
	_snapToOrthogonal(&p);
	_positionMagnifyRect(&screenFrame, &p);

	screen.ReadBitmap(_screenSnip, false, &_magnifyRect);
	
	float dx = 20, dy = 20;
	
	//we want to position the window, but not to move it offscreen
	if ((p.x + dx + swatchBounds.Width()) > screenFrame.Width())
	{
		dx = - swatchBounds.Width() - dx;
	}

	if ((p.y + dy + swatchBounds.Height()) > screenFrame.Height())
	{
		dy = - swatchBounds.Height() - dy;
	}	

	Window()->MoveTo(BPoint(p.x + dx, p.y + dy));
		
	_mouseLoc = cleanPoint;
	
	Invalidate();
}

void MagnifyView::SetMagnification(double mag)
{
	_magnification = mag;
	if (_magnification < 1) mag = 1;

	BRect b(Bounds());
	float w = b.Width();
	float h = b.Height();
	
	_magnifyRect = b;
	_magnifyRect.right = _magnifyRect.left + (w/_magnification);
	_magnifyRect.bottom = _magnifyRect.top + (h/_magnification);


	if (_screenSnip) delete _screenSnip;
	_screenSnip = new BBitmap(_magnifyRect, B_RGB32);
}

double MagnifyView::GetMagnification()
{
	return _magnification;
}

void MagnifyView::_positionMagnifyRect(BRect *screenFrame, BPoint *mouseLocation)
{
	//position _magnifyRect with mouseLocation in center, but adjust for edges of screen
	//assumes _magnifyRect has already had a height and width set when SetMagnification is called.
	float magWidth = _magnifyRect.Width();
	float magHeight = _magnifyRect.Height();
	
	_magnifyRect.top = mouseLocation->y - magHeight / 2;
	_magnifyRect.bottom = _magnifyRect.top + magHeight;
	_magnifyRect.left = mouseLocation->x - magWidth / 2;
	_magnifyRect.right = _magnifyRect.left + magWidth;

	if (_magnifyRect.left < 0)
		_magnifyRect.OffsetBy(-_magnifyRect.left, 0);
		
	if (_magnifyRect.top < 0)
		_magnifyRect.OffsetBy(0, -_magnifyRect.top);
		
	if (_magnifyRect.right > screenFrame->right)
		_magnifyRect.OffsetBy(screenFrame->right - _magnifyRect.right, 0);
		
	if (_magnifyRect.bottom > screenFrame->bottom)
		_magnifyRect.OffsetBy(0, screenFrame->bottom - _magnifyRect.bottom);
	
}

void MagnifyView::_checkForShiftKey(BPoint *p)
{
	if (modifiers() & B_SHIFT_KEY)
	{
		//we only want to do this once, when the shift key is first pressed
		if (!_shiftDown)
		{
			_shiftDown = true;
			//record the position of the cursor, so we can snap to the axes later
			_shiftDownAt = *p;
		}
	}
	else
		_shiftDown = false;
}

void MagnifyView::_snapToOrthogonal(BPoint *p)
{
	if (_shiftDown)
	{
		//we know _shiftDownAt is set when the shift key is first pressed.
		//so we want to snap 'p' to the vertical or horizontal axis it is closest to
		//to determine which axis p is closest to, a little trig:
		
		float dy = (p->y - _shiftDownAt.y);
		float dx = (p->x - _shiftDownAt.x);
		
		dy = _ABS(dy);
		dx = _ABS(dx);
		
		if (dy > dx) //line is more vertical
		{
			p->x = _shiftDownAt.x;
		}
		else //line is more horizontal
		{
			p->y = _shiftDownAt.y;
		}
	}
}
