#include "zlayout.h"

ZFlowLayout::ZFlowLayout(long alignment, float height=32, float width=0, int hPad = 2, int vPad = 2, int children=36)
	: ZLayoutManager(children)
{
	_align = alignment;
	_fixedHeight = height;
	_fixedWidth = width;
	
	_hPad = hPad;
	_vPad = vPad;
	
	SetJustified(false);
	_preferredWidth = _preferredHeight = 0;

	_rowList = new int[children];
}

ZFlowLayout::~ZFlowLayout()
{
	delete _rowList;
}

void ZFlowLayout::Draw(ZLayoutView *v, float w, float h)
{
	//right now this does nothing more than draw a frame around the contents. Very simple.
	rgb_color c = tint_color(v->ViewColor(), B_LIGHTEN_1_TINT);
	
	v->SetPenSize(0);
	v->SetHighColor(c);
	

	
	if (_align == LEFT)
		v->StrokeRoundRect(BRect(0,0, _preferredWidth, _preferredHeight), 3, 3);
	else if (_align == RIGHT)
		v->StrokeRoundRect(BRect(w - _preferredWidth,0, w, _preferredHeight), 3, 3);
}

void ZFlowLayout::Layout(float w, float h)
{
	//process is like laying out text
	//first element inserted is first element of first row. Elements are placed from left to right
	//or right to left depending on alignment, wrapping if too wide for remaining space.
	//proportions are clipped to fixed height & width; to allow content-sensitive 
	//proportioning, set _fixedHeight or _fixedWidth to 0
	
	#ifdef DEBUG_FLOW
	printf("ZFlowLayout::Layout(%f, %f)\n", w, h);
	#endif
	
	float x = ((_align == LEFT) ? 0 : w), y = 0;
	float pw, ph, fpw, fph;	
	int children = Size(), rows = 0, childIndex = 0;
	BRect f;
	
	 //Frame Preferred Width/Height -- we precalculate the preferred width & height for the ZLayoutView
	 //by recording the smallest rectangle which will hold the contents after being layed out at this size
	fpw = fph = 0;
	
	float *rowWidths = new float[children + 1]; //you can't have more rows than children

	if (_justify)
		for (int i = 0; i < children; i++) _rowList[i] = -1;

	for (int i = 0; i < children; i++)
	{
		BView *child = ChildAt(i);
		if (child)
		{		
			child->GetPreferredSize(&pw, &ph);
			
			if (_fixedWidth) pw = _fixedWidth;

				
			switch (_align)
			{
				case LEFT:
				{
					//wrap if this kid will go over the edge, and if it's not the first
					if (((x + pw + _hPad) > w) && childIndex)
					{
						rowWidths[rows++] = x;

						x = 0;
						y += _fixedHeight + _vPad;
						
					}

					f.left = x;
					f.top = y;
					f.right = f.left + pw;
					f.bottom = f.top + _fixedHeight;					
					
					x += pw + _hPad;

					if (f.right > fpw) fpw = f.right;
					if (f.bottom > fph) fph = f.bottom;			

					break;
				}

				case RIGHT:
				{
					//wrap if this kid will go over the edge, and if it's not the first
					if (((x - (pw + _hPad)) < 0) && childIndex)
					{
						rowWidths[rows++] = w - x;
						
						x = w;
						y += _fixedHeight + _vPad;
					}

					f.right = x;
					f.top = y;
					f.left = f.right - pw;
					f.bottom = f.top + _fixedHeight;
					
					x -= pw + _hPad;

					if ((w - f.left) > fpw) fpw = (w - f.left);
					if (f.bottom > fph) fph = f.bottom;
			
					break;
				}			
			
			}
			
			_rowList[i] = rows; //asign row index to this child
			
			if (child->Frame() != f) //no point setting child frame unless it needs to be changed.
			{
				child->MoveTo(f.left, f.top);
				child->ResizeTo(f.Width(), f.Height());
			}
			
			childIndex++;
		}
	}
	
	//now that everyody's all happily layed out per preferred width, let's justify, row per row
	if (!_fixedWidth && _justify)
	{
		BView *v;
		
		for (int r = 0; r < rows; r++)
		{
			float scaleUpX = w / rowWidths[r]; //this is how much to horizontally scale this row to fit the width
			float dx = w - rowWidths[r]; //align_right needs to be offset to the left
		
			for (int i = 0; i < children; i++)
			{					
				v = ChildAt(i);
				if (v && (_rowList[i] == r))
				{
					if (_align == LEFT)
					{
						BRect f = v->Frame();
						f.left *= scaleUpX;
						f.right *= scaleUpX;
						v->MoveTo(f.left, f.top);
						v->ResizeTo(f.Width(), f.Height());
					}
					else if (_align == RIGHT)
					{
						BRect f = v->Frame();
						
						f.left -= dx;
						f.right -=dx;
						
						f.left *= scaleUpX ;
						f.right *= scaleUpX;
						v->MoveTo(f.left, f.top);
						v->ResizeTo(f.Width(), f.Height());					
					}
				}
			}
		}
	}
	
	
	if (!fph && !fpw)
	{
		_preferredWidth = 0;
		_preferredHeight = 0;
	}
	else
	{
		_preferredWidth = fpw + _hPad;
		_preferredHeight = fph + _vPad;
	}
	
	delete [] rowWidths;	
}

void ZFlowLayout::GetPreferredSize(float *w, float *h, float currentWidth, float currentHeight)
{ 
	//flow layout doesn't try to compute the optimum rectangle for
	//laying out the contents, rather, during the layout phase, it
	//figures out the minimum rectangle which contains its contents
	//at its current size. Then it determines if it isn't wide enough, if so, it returns
	//a new width which is. Otherwise, it returns the current width
	
	if (_preferredWidth < currentWidth)	
		*w = _preferredWidth;
	else
		*w = currentWidth;
		
	*h = _preferredHeight;
}

void ZFlowLayout::SetFixedWidth(float w)
{
	_fixedWidth = w;
}

float ZFlowLayout::FixedWidth()
{
	return _fixedWidth;
}

void ZFlowLayout::SetFixedHeight(float h = 0)
{
	_fixedHeight = h;
}

float ZFlowLayout::FixedHeight()
{
	return _fixedHeight;
}

void ZFlowLayout::SetJustified(bool tf)
{
	//justification will stretch elements which fit on a row to take up entire row
	//will not compromise any (eg, will not shrink any components smaller than preferred sizes)
	_justify = tf;
}

bool ZFlowLayout::Justified()
{
	return _justify;
}

void ZFlowLayout::SetAlignment(long a)
{
	_align = a;
}

long ZFlowLayout::Alignment()
{
	return _align;
}
