#include "zlayout.h"

ZGridLayout::ZGridLayout(int32 cols, int32 rows, int hPad = 2, int vPad = 2)
	: ZLayoutManager((cols * rows))
{
	_rows = rows;
	_cols = cols;
	
	_hPad = (float) hPad;
	_vPad = (float) vPad;
	
	_cellHeight = 0;
	_cellWidth = 0;
}

ZGridLayout::~ZGridLayout(){;}

void ZGridLayout::SetCellHeight(float h){
	_cellHeight = h;
}

void ZGridLayout::SetCellWidth(float w){
	_cellWidth = w;
}

float ZGridLayout::CellHeight() { return _cellHeight; }
float ZGridLayout::CellWidth() { return _cellWidth; }

void ZGridLayout::Draw(ZLayoutView *v, float w, float h)
{
	rgb_color c = tint_color(v->ViewColor(), B_LIGHTEN_1_TINT);
	
	v->SetPenSize(0);
	BRect bnds = BRect(0, 0, w, h);
	v->SetHighColor(c);
	v->StrokeRoundRect(bnds, 3, 3);
		
	float dx, dy;
	dx = (w / _cols);
	dy = (h / _rows);

	if ((dx > _cellWidth) && (_cellWidth > 0)) dx = _cellWidth;
	if ((dy > _cellHeight) && (_cellHeight > 0)) dy = _cellHeight;

	BPoint p, q;
	
	v->BeginLineArray(_cols + _rows + 2);
	
	p.x = 0; q.x = w;
	for (int row = 1; row < _rows; row++)
	{
		p.y = q.y = row * dy;
		v->AddLine(p, q, c);
	}
	
	p.y = 0; q.y = h;
	for (int col = 1; col < _cols; col++)
	{
		p.x = q.x = col * dx;
		v->AddLine(p, q, c);
	}
	
	v->EndLineArray();
}

void ZGridLayout::GetPreferredSize(float *width, float *height, float currentWidth, float currentHeight)
{

	
	//two steps -- first find maximum height & width of all elements,
	float maxWidth = 0, maxHeight = 0;
	float tw, th;
	
	for(int i = 0; i < Size(); i++)
	{
		BView *child = ChildAt(i);
		if (child)
		{	
			child->GetPreferredSize(&tw, &th);
			
			if (tw > maxWidth) maxWidth = tw;
			if (th > maxHeight) maxHeight = th;				
		}
	}	

	//second step -- if _cellWidth or _cellHeight are non-zero but less than max values,
	//set max values to them
	if ((_cellWidth > 0) && (_cellWidth < maxWidth)) maxWidth = _cellWidth;
	if ((_cellHeight > 0) && (_cellHeight < maxHeight)) maxHeight = _cellHeight;

	*width = _cols * (maxWidth + _hPad);
	*height = _rows * (maxHeight + _vPad);
}

void ZGridLayout::Layout(float w, float h){
		
	#ifdef DEBUG_GRID	
	printf("ZGridLayout::Layout()\n");
	#endif

	int whichView = 0;

	float dx, dy;

	dx = (w / _cols);
	dy = (h / _rows);
	
	#ifdef DEBUG_GRID
	printf("\tdx = %f; dy = %f\t_cellWidth = %f; _cellHeight = %f\n", dx, dy, _cellWidth, _cellHeight);
	#endif
	
	if ((dx > _cellWidth) && (_cellWidth > 0)) dx = _cellWidth;
	if ((dy > _cellHeight) && (_cellHeight > 0)) dy = _cellHeight;
	
	#ifdef DEBUG_GRID
	printf("\tdx = %f; dy = %f\t_cellWidth = %f; _cellHeight = %f\n", dx, dy, _cellWidth, _cellHeight);
	#endif

	
	float hP = _hPad / 2.0;
	float vP = _vPad / 2.0;
	float left, top, width, height;

	width = dx - 1;
	height = dy - 1;

	for (int row = 0; row < _rows; row++){
		for (int col = 0; col < _cols; col++){
			
			left = (float)col * dx + 1;
			top = (float)row * dy + 1;
						
			#ifdef DEBUG_GRID
			printf("\tView %d, (%f, %f, %f, %f)\n", whichView, left, top, width, height);
			#endif
			
			BView *view = ChildAt(whichView);
			
			if (view){
				view->ResizeTo(width - _hPad, height - _vPad);
				view->MoveTo(left + hP, top + vP);
			}
			else{
				#ifdef DEBUG_GRID
				printf("No view at index %d to layout.\n", whichView);
				#endif
			}
			
			whichView++;		
		}
	}

}
