#include "zlayout.h"

ZHorizontalLayout::ZHorizontalLayout(int hPad = 2, int vPad = 2, int children = 36)
	:ZLayoutManager(children)
{
	_spring = -1;
	_hPad = hPad;
	_vPad = vPad;

	_fixedWidth = 0;
	_fixedHeight = 0;
	
	_childRects = new BRect[children];
	SetJustified(true);
}

ZHorizontalLayout::~ZHorizontalLayout(){ delete [] _childRects;}

void ZHorizontalLayout::PutSpring(int where)
{
	if (where > CountChildren()) where = CountChildren() - 1;
	if (where < 0) where = -1;

	_spring = where;
}

void ZHorizontalLayout::RemoveSpring()
{
	_spring = -1;
}

void ZHorizontalLayout::SetJustified(bool j)
{
	_justify = j;
}

bool ZHorizontalLayout::Justified()
{
	return _justify;
}

void ZHorizontalLayout::SetFixedWidth(float w = 0)
{
	_fixedWidth = w;
}

float ZHorizontalLayout::FixedWidth()
{
	return _fixedWidth;
}

void ZHorizontalLayout::SetFixedHeight(float h)
{
	_fixedHeight = h;
}

float ZHorizontalLayout::FixedHeight()
{
	return _fixedHeight;
}

void ZHorizontalLayout::Layout(float w, float h)
{
	//first, we lay out all the children into _childRects
	//then we deternime where to put the spring (if at all) and reposition the left from the right
	//then we put the children into their corresponding positions (based on the rects)
	
	int size = Size();

	//calculate where children would go sans a spring
	int rectIndex = 0;
	for (int i = 0; i < size; i++)
	{
		BView *child = ChildAt(i);
		if (child)
		{
			//much to decide here
			//first determine how big child would like to be
			//if we've set a preferred width or height, ovveride GetPreferredSize
			//now, since layout is horizontal, if layout is justified, we need to stretch each component to bottom edge
			//but only if preferred heights aren't set.
		
			float prefWidth, prefHeight;
			child->GetPreferredSize(&prefWidth, &prefHeight);
			if (_fixedHeight) prefHeight = _fixedHeight;
			else if (_justify) prefHeight = h - 2 *_vPad;
					
			if (_fixedWidth) prefWidth = _fixedWidth;

			//special case for first child in row, position relative to upper left of view (with padding)				
			if (rectIndex == 0)
			{
				_childRects[rectIndex].Set(_hPad, _vPad, _hPad + prefWidth, _vPad + prefHeight);
			}
			else //position the n relative to (n - 1)
			{
				BRect *lastRect = &_childRects[rectIndex - 1];
				_childRects[rectIndex].Set(lastRect->right + _hPad, _vPad, lastRect->right + _hPad + prefWidth, _vPad + prefHeight);			
			}
			
			rectIndex++;
		}
	}
	
	//if there's a spring, move _childRects at and after _spring to align to right-hand boundary
	if (_spring >= 0)
	{
		float dx = 0;

		//find .right of last child		
		for (int i = _spring; i < size; i++)
		{
			if (ChildAt(i))
				dx = _childRects[i].right;
		}
		
		//how far must they be translated?
		dx = w - dx - _hPad;
		
		if (dx > 0)
		{
			for (int i = _spring; i < size; i++)
			{
				if (ChildAt(i)) _childRects[i].OffsetBy(dx, 0);		
			}
		}
	}
	
	//position children per what we've calculated.
	for (int i = 0; i < size; i++)
	{
		BView *child = ChildAt(i);
		if (child)
		{
			//don't resize/move if it's already at the right size/position
			if (_childRects[i] != child->Frame())
			{
				child->MoveTo(_childRects[i].LeftTop());
				child->ResizeTo(_childRects[i].Width(), _childRects[i].Height());
			}
		}	
	}
}

void ZHorizontalLayout::GetPreferredSize(float *w, float *h, float currentWidth, float currentHeight)
{
	float maxChildHeight = 0, summedWidths = _hPad;
	float tw, th;
	int size = Size();

	//first, sum up the preferred widths and find the maximum height of the children	
	//we're going to ignore the spring, if there is one. We're only concerned with the contents
	for(int i = 0; i < size; i++)
	{
		BView *child = ChildAt(i);
		if (child)
		{	
			child->GetPreferredSize(&tw, &th);
			
			if (th > maxChildHeight) maxChildHeight = th;
			summedWidths += tw + _hPad;
		}
	}
	
	maxChildHeight += 2 * _vPad;

	//if user has chosen a preferred width or height, set it so
	if (_fixedHeight != 0) maxChildHeight = _fixedHeight + 2*_vPad;
	if (_fixedWidth != 0) summedWidths = (_fixedWidth + _hPad) * CountChildren() + _hPad;

	*w = summedWidths;
	*h = maxChildHeight;
}