#include "zlayout.h"

ZLayoutView::ZLayoutView(BRect frame, const char *name, ZLayoutManager *lm, uint32 resize = B_FOLLOW_LEFT | B_FOLLOW_TOP, uint32 flags = 0)
	: BView(frame, name, resize, flags | B_FRAME_EVENTS /*| B_DRAW_ON_CHILDREN | B_WILL_DRAW | B_FULL_UPDATE_ON_RESIZE*/)
{
	_layoutManager = lm;
	_windowZoomer = NULL;
	SetTopmost(false);	
	DrawLayout(false);		
	
	//when these are set to -1, GetPreferredSize(...) will calculate preferred
	//size based on contents of ZLayoutView. On the other hand, if these have a
	//different value, as set by SetPreferredSize(...), GetPreferredSize will return
	//these values. Useful for if you have a view which should remain a certain size
	//while you still want the rest of the layout to be smartly resizeable.
	_preferredWidth = -1;
	_preferredHeight = -1;	
}

ZLayoutView::~ZLayoutView()
{
	//deletion of manager doesn't delete views -- we leave this to BView::~BView
	//layout manager just moves them around.
	delete _layoutManager;
}

ZLayoutManager * ZLayoutView::LayoutManager()
{
	return _layoutManager;
}

void ZLayoutView::InvalidateLayout()
{
	//no point if we're not attached yet to window
	if (!Window()) return;

	#ifdef DEBUG_LV
	printf("ZLayoutView %s::InvalidateLayout()\n", Name());
	#endif

	//go through all the children and Invlidate their layout, if they're ZLayoutViews
	BView *v;
	for (int i = 0; i < _layoutManager->Size(); i++)
	{
		v = _layoutManager->ChildAt(i);
		if (v && (typeid(*v) == typeid(ZLayoutView)))
		{
			((ZLayoutView *)v)->LayoutManager()->Layout(v->Bounds().Width(), v->Bounds().Height());					
		}
	}	
	
	//finally do our layout with our contents
	_layoutManager->Layout(Bounds().Width(), Bounds().Height());	

}

void ZLayoutView::setFlags(BView *v){
//dynamic resizing requires that B_FULL_UPDATE_ON_RESIZE be set to look 'good'
//as well as B_FOLLOW_NONE such that Be's internal layout management is overrided
//problem is, the full update is slooow. I'm looking into locking the appserver
//from drawing as often. perhaps with a call to Window()->OpenViewTransaction... which
//although the BeBook lists it, doesn't seem to exist.

	if (typeid(*v) == typeid(BTextControl))
		{
			//unfortunately class BTextControl must be handled differently than the others
			//while we're at it, might as well automagically call set_divider to fit label
			//since it's a pain to do all the time.
			BTextControl *tc = (BTextControl *) v;
			tc->TextView()->SetFlags(tc->TextView()->Flags() | B_FULL_UPDATE_ON_RESIZE);
			
			tc->SetDivider(tc->StringWidth(tc->Label()) + 10);
		}
	else
		{
			v->SetFlags(v->Flags() | B_FULL_UPDATE_ON_RESIZE);	
		}


	v->SetResizingMode(B_FOLLOW_NONE);
}


void ZLayoutView::AddChild(BView *v, BView *sibling = NULL)
{
	BView::AddChild(v, sibling);

	setFlags(v);
	
	#ifdef DEBUG_LV
	printf("ZLayoutView %s::AddChild() : %s\n", Name(), v->Name());
	#endif
	
	_layoutManager->AddChild(v);

	if (Parent() && (typeid(*Parent()) == typeid(ZLayoutView)))
		((ZLayoutView *)Parent())->InvalidateLayout();
	else
		InvalidateLayout();
}

void ZLayoutView::AddChildAt(BView *v, int32 where = 0)
{
	BView::AddChild(v);

	setFlags(v);

	#ifdef DEBUG_LV
	printf("ZLayoutView %s::AddChildAt() : %s\n", Name(), v->Name());
	#endif
	
	_layoutManager->AddChildAt(v, where);

	if (Parent() && (typeid(*Parent()) == typeid(ZLayoutView)))
		((ZLayoutView *)Parent())->InvalidateLayout();
	else
		InvalidateLayout();
}

void ZLayoutView::AddChildWithConstraint(BView *v, ZConstraint *constraint)
{
	BView::AddChild(v);

	setFlags(v);

	#ifdef DEBUG_LV
	printf("ZLayoutView %s::AddChildWithConstraint() : %s\n", Name(), v->Name());
	#endif
	
	_layoutManager->AddChildWithConstraint(v, constraint);

	if (Parent() && (typeid(*Parent()) == typeid(ZLayoutView)))
		((ZLayoutView *)Parent())->InvalidateLayout();
	else
		InvalidateLayout();
}

bool ZLayoutView::RemoveChild(BView *v)
{
	_layoutManager->RemoveChild(v);
	BView::RemoveChild(v);

	#ifdef DEBUG_LV
	printf("ZLayoutView %s::RemoveChild() : %s\n", Name(), v->Name());
	#endif


	if (Parent() && (typeid(*Parent()) == typeid(ZLayoutView)))
		((ZLayoutView *)Parent())->InvalidateLayout();
	else
		InvalidateLayout();

	return true;
}

void ZLayoutView::AttachedToWindow(){
//set to bgcolor of parent; if no parent, set to default BeOS grey because
//default white color is ass-ugly ;)
	if (Parent())
		SetViewColor(Parent()->ViewColor());
	else
		SetViewColor(ui_color(B_PANEL_BACKGROUND_COLOR));
	
	//if the view constitues the bounds rect of the window, make it topmost	
	SetTopmost((Bounds() == Window()->Bounds()));
	BView::AttachedToWindow();
}

void ZLayoutView::AllAttached()
{
	//all children have received attachment notification
	InvalidateLayout();
	BView::AllAttached();
}


void ZLayoutView::FrameResized(float width, float height)
{
	#ifdef DEBUG_LV
	printf("ZLayoutView %s::FrameResized(%f, %f)\n", Name(), width, height);
	#endif
	
	if (IsTopmost() && Window()->Lock())
	{
		Window()->BeginViewTransaction();
		_layoutManager->Layout(width, height);
	
		Window()->EndViewTransaction();		
		Window()->Unlock();
	}
	else
		_layoutManager->Layout(width, height);		
	
}

void ZLayoutView::DrawAfterChildren(BRect r)
{
	if (_layoutManager->ShouldDraw() && _layoutManager->CountChildren()) 
	{
		BRect b(Bounds());
		_layoutManager->Draw(this, b.Width(), b.Height());
	}
}

void ZLayoutView::SetPreferredSize(float width, float height)
{
	_preferredWidth = width;
	_preferredHeight = height;
	
	ResizeToPreferred();
}

void ZLayoutView::GetPreferredSize(float *width, float *height)
{
	#ifdef DEBUG_LV
	printf("ZLayoutView %s::GetPreferredSize()\n", Name());
	#endif

	float w, h;
	InvalidateLayout();	
	_layoutManager->GetPreferredSize(&w, &h, Bounds().Width(), Bounds().Height());
	
	//let user-set width and or height take precedence
	if (_preferredWidth != -1) w = _preferredWidth;
	if (_preferredHeight != -1) h = _preferredHeight;


	//default layout manager implentation will return -1 for both, signalling
	//that there is no preference for preferred size
	//but derivations of ZLayoutManager may return -1 
	//for just one proportion and some 'actual' value for the other
	if (w == -1) w = Bounds().Width();
	if (h == -1) h = Bounds().Height();
	
	*width = w;
	*height = h;
}

void ZLayoutView::ResizeToPreferred()
{
	#ifdef DEBUG_LV
	printf("ZLayoutView %s :: ResizeToPreferred()\n", Name());
	#endif

	float w, h;
	GetPreferredSize(&w, &h);

	#ifdef DEBUG_LV
	printf("\tResizing to %f, %f\n", w, h);
	#endif

	ResizeTo(w, h);
			
	if (IsTopmost())
	{
		//unset the following mode
		SetResizingMode(B_FOLLOW_NONE);

		//we don't want to window freaking out and becoming too big for the screen
		BRect windowFrame = Window()->Frame();
		BScreen screen;

		if (w > screen.Frame().Width()) w = screen.Frame().Width();
		if (h > screen.Frame().Height()) h = screen.Frame().Height();
		
		if (windowFrame.left + w > screen.Frame().Width())
		{
			Window()->MoveTo(screen.Frame().right - w, windowFrame.top);
		}
		
		if (h > screen.Frame().Height())
		{
			Window()->MoveTo(windowFrame.left, screen.Frame().bottom - h);
		}
				
		Window()->ResizeTo(w, h);	
		SetResizingMode(B_FOLLOW_ALL_SIDES);
	}
}

void ZLayoutView::DrawLayout(bool tf)
{
	if (tf)
	{
		SetFlags(Flags() | B_WILL_DRAW | B_FULL_UPDATE_ON_RESIZE | B_DRAW_ON_CHILDREN);
	}
	else
	{
		//undo those flags responsible for drawing -- a performance increase during resize.
		SetFlags(Flags() & ~(B_WILL_DRAW | B_FULL_UPDATE_ON_RESIZE | B_DRAW_ON_CHILDREN));
	}
	
	_layoutManager->SetDrawing(tf);
	Invalidate();

}

void ZLayoutView::SetTopmost(bool tf)
{ 
	#ifdef DEBUG_LV
	printf("ZLayoutView %s::SetTopmost()\n", Name());
	#endif

	_topmost = tf;


	//set up a ZWindowZoomer to intercept B_ZOOM messages
	//A ZWindowZoomer calls ResizeToPreferred on the view
	//its initialized with when it receives a B_ZOOM message
	if (_topmost)
	{
		//delete old window zoomer (I don't think this'll ever occur, 
		//but it just might prevent a minor memory leak)
		if (_windowZoomer)
		{
			Window()->RemoveCommonFilter(_windowZoomer);
			delete _windowZoomer;
			_windowZoomer = NULL;
		}
		_windowZoomer = new ZWindowZoomer(this);
		Window()->AddCommonFilter(_windowZoomer);	
	}
	else
	{
		if (_windowZoomer)
		{
			Window()->RemoveCommonFilter(_windowZoomer);
			delete _windowZoomer;
			_windowZoomer = NULL;
		}	
	}			
}

bool ZLayoutView::IsTopmost(){ return _topmost; }
