#include "zlayout.h"

#include <Cursor.h>
#include <Application.h>
#include <Resources.h>

#define ZSPLIT_PANE_STEP_CLOSE 'zSCD'
#define ZSPLIT_PANE_STEP_OPEN 'zSOP'

//***************************************************************
//	Just a function for drawing a raised, sunken, or faded flush rectangle
//	Provided as convenience function for drawing the 'grippers' for the split pane
//***************************************************************
void _draw3DRect(BView *v, BRect r, rgb_color bg, rgb_color light, rgb_color dark, long frameStyle)
{
	rgb_color originalColor = v->HighColor();

	v->SetPenSize(0);
	v->SetHighColor(bg);
	v->FillRect(r);

	switch(frameStyle)
	{
		case ZSplitPane::RAISED:		
		{
			v->SetHighColor(dark);
			v->StrokeLine(r.RightTop(), r.RightBottom()); v->StrokeLine(r.LeftBottom(), r.RightBottom());
		
			v->SetHighColor(light);
			v->StrokeLine(r.LeftTop(), r.RightTop()); v->StrokeLine(r.LeftTop(), r.LeftBottom());
			break;
		}

		case ZSplitPane::SUNKEN:
		{
			v->SetHighColor(light);
			v->StrokeLine(r.RightTop(), r.RightBottom()); v->StrokeLine(r.LeftBottom(), r.RightBottom());
		
			v->SetHighColor(dark);
			v->StrokeLine(r.LeftTop(), r.RightTop()); v->StrokeLine(r.LeftTop(), r.LeftBottom());
			break;
		}
	
		case ZSplitPane::FLUSH:
		{
			drawing_mode oldDrawingMode = v->DrawingMode();
			v->SetDrawingMode(B_OP_BLEND);

			r.OffsetBy(1, 1);			
			
			v->SetHighColor(dark);
			v->StrokeRect(r);
		
			r.OffsetBy(1, 1);
			v->SetHighColor(light);
			v->StrokeRect(r);
			
			v->SetDrawingMode(oldDrawingMode);
			break;
		}
	}
	
	v->SetHighColor(originalColor);
}


//***************************************************************


ZSplitPane::ZSplitPane(BRect frame, const char *name, long orientation = HORIZONTAL, uint32 resize = B_FOLLOW_ALL_SIDES, uint32 flags = 0)
	: BView(frame, name, resize, flags | B_WILL_DRAW | B_FRAME_EVENTS)
{
	_orientation = orientation;
	_firstChild = _secondChild = NULL;
	_mouseDown = false;
	
	_snapped = false;
	_splitterWidth = 9;
	_snapDistance = 20;
	
	_lastWidth = frame.right;
	_lastHeight = frame.bottom;

	//set default following mode
	SetFollowingMode(FOLLOW_FIRST);

	_hCursorLR = _hCursorL = _hCursorR = _vCursorTB = _vCursorT = _vCursorB = NULL;
	BResources *resources = be_app->AppResources();
	size_t outSize;
	const void *data;
	
	data = resources->LoadResource('CURS', "h_cursor_left_right", &outSize);
	if (data) _hCursorLR = new BCursor(data);
	else printf("ZSplitPane::ZSplitPane()\tUnable to load left_right cursor\n");

	data = resources->LoadResource('CURS', "h_cursor_left", &outSize);
	if (data) _hCursorL = new BCursor(data);
	else printf("ZSplitPane::ZSplitPane()\tUnable to load left cursor\n");

	data = resources->LoadResource('CURS', "h_cursor_right", &outSize);
	if (data) _hCursorR = new BCursor(data);
	else printf("ZSplitPane::ZSplitPane()\tUnable to load right cursor\n");

	data = resources->LoadResource('CURS', "v_cursor_top_bottom", &outSize);
	if (data) _vCursorTB = new BCursor(data);
	else printf("ZSplitPane::ZSplitPane()\tUnable to load top_bottom cursor\n");

	data = resources->LoadResource('CURS', "v_cursor_top", &outSize);
	if (data) _vCursorT = new BCursor(data);
	else printf("ZSplitPane::ZSplitPane()\tUnable to load top cursor\n");

	data = resources->LoadResource('CURS', "v_cursor_bottom", &outSize);
	if (data) _vCursorB = new BCursor(data);
	else printf("ZSplitPane::ZSplitPane()\tUnable to load bottom cursor\n");
	
	_messageRunner = NULL;


	SetSplitPosition((_orientation == HORIZONTAL) ? Bounds().Width() / 2 : Bounds().Height() / 2);
}

ZSplitPane::~ZSplitPane()
{
	if (_hCursorLR) delete _hCursorLR;
	if (_hCursorL) delete _hCursorL;
	if (_hCursorR) delete _hCursorR;
	if (_vCursorTB) delete _vCursorTB;
	if (_vCursorT) delete _vCursorT;
	if (_vCursorB) delete _vCursorB;
	
	_stopAnimation();
}

void ZSplitPane::AddFirstChild(BView *child)
{
	if (!child) return;
	if (_firstChild) RemoveChild(_firstChild);

	child->SetResizingMode(B_FOLLOW_NONE);	
	_firstChild = child;
	AddChild(_firstChild);
	_doLayout();
}

void ZSplitPane::AddSecondChild(BView *child)
{
	if (!child) return;
	if (_secondChild) RemoveChild(_secondChild);
	
	_secondChild = child;
	child->SetResizingMode(B_FOLLOW_NONE);	
	AddChild(_secondChild);
	_doLayout();
}

void ZSplitPane::RemoveFirstChild()
{
	if (_firstChild){
		RemoveChild(_firstChild);
		_firstChild = NULL;
	}
	
	_doLayout();
}

void ZSplitPane::RemoveSecondChild()
{
	if (_secondChild){
		RemoveChild(_secondChild);
		_secondChild = NULL;
	}
	
	_doLayout();
}

void ZSplitPane::AttachedToWindow()
{
	SetViewColor(ui_color(B_PANEL_BACKGROUND_COLOR));

	//to insure MouseMoved() doesn't stop up the window thread
	Window()->SetFlags( Window()->Flags() | B_ASYNCHRONOUS_CONTROLS);
}

void ZSplitPane::GetPreferredSize(float *width, float *height)
{
	float fcW = 0, fcH = 0, scW = 0, scH = 0;
	if (_firstChild) _firstChild->GetPreferredSize(&fcW, &fcH);
	if (_secondChild) _secondChild->GetPreferredSize(&scW, &scH);
	float w = Bounds().Width(), h = Bounds().Height();

	//discard preferred dimensions of snapped-shut child
	if (_snapped && (_snappedAt == SNAP_FIRST))
	{
		fcW = fcH = 0;
	}

	if (_snapped && (_snappedAt == SNAP_SECOND))
	{
		scW = scH = 0;
	}


	if (_orientation == HORIZONTAL)
	{		
		if (_follow == FOLLOW_FIRST)				
			w = _splitterRect.right + scW + 1;
		else
			w = (Bounds().right - _splitterRect.left) + fcW;
			
		h = (fcH > scH) ? fcH : scH;	
	}
	else
	{
		if (_follow == FOLLOW_FIRST)
			h = _splitterRect.bottom + scH + 1;
		else
			h = (Bounds().bottom - _splitterRect.top) + fcH;
			
		w = (fcW > scW) ? fcW : scW;	
	}
	
	*width = w;
	*height = h;
}


void ZSplitPane::FrameResized(float width, float height)
{
	//resize the frame of the splitter
	if (_orientation == HORIZONTAL)
	{
		_splitterRect.bottom = height;
		_preSnapSplitterRect.bottom = height;
	}
	else
	{
		_splitterRect.right = width;
		_preSnapSplitterRect.right = width;
	}

	//now move splitter if the following mode is SECOND
	if (_follow == FOLLOW_SECOND)
	{
		if (_orientation == HORIZONTAL)
		{
			_splitterRect.OffsetBy(width - _lastWidth, 0);
			_preSnapSplitterRect.OffsetBy(width - _lastWidth, 0);
		}
		else
		{
			_splitterRect.OffsetBy(0, height - _lastHeight);
			_preSnapSplitterRect.OffsetBy(0, height - _lastHeight);
		}
	
	}

	//remember last position
	_lastWidth = width;
	_lastHeight = height;

	_doLayout();
}

void ZSplitPane::MouseDown(BPoint loc)
{
	if (_splitterRect.Contains(loc))
	{
		_mouseDown = true;
		_mouseDownPosition = loc;
		SetEventMask(B_POINTER_EVENTS, B_NO_POINTER_HISTORY);
		Invalidate();

		//we want to remember the last splitter position		
		if (!_snapped)
			_preSnapSplitterRect = _splitterRect;
	}
}

void ZSplitPane::MouseMoved(BPoint loc, uint32 transit, const BMessage *msg)
{

	//set a cursor (if loaded from resources)
	if (transit == B_ENTERED_VIEW)
	{	
		if (_orientation == HORIZONTAL)
		{
			if (!_snapped)
			{
				if (_hCursorLR) SetViewCursor(_hCursorLR);
			}
			else
			{				
				if (_snappedAt == SNAP_FIRST)
				{ //the first pane is snapped shut, so show only the right cursor
					if (_hCursorR) SetViewCursor(_hCursorR);
				}
				else if (_snappedAt == SNAP_SECOND)
				{//the second pane is snapped shut, so show the left cursor arrow
					if (_hCursorL) SetViewCursor(_hCursorL);
				}
			}
		}
		else
		{
			if (!_snapped)
			{
				if (_vCursorTB) SetViewCursor(_vCursorTB);
			}
			else
			{
				if (_snappedAt == SNAP_FIRST)
				{ //the first pane is snapped shut, so show only the down cursor
					if (_vCursorB) SetViewCursor(_vCursorB);
				}
				else if (_snappedAt == SNAP_SECOND)
				{//the second pane is snapped shut, so show the up cursor arrow
					if (_vCursorT) SetViewCursor(_vCursorT);
				}
			}
		}
	}
	else if (transit == B_EXITED_VIEW)
	{
		//go back to system default cursor
		SetViewCursor(B_CURSOR_SYSTEM_DEFAULT);
	}

	if (!_mouseDown) return;

	//now do the actual work of setting the splitter position
	if (_orientation == HORIZONTAL)
	{
		if (loc.x < 0) loc.x = 0;
		SetSplitPosition(loc.x);
	}
	else
	{
		if (loc.y < 0) loc.y = 0;
		SetSplitPosition(loc.y);		
	}
}

void ZSplitPane::MouseUp(BPoint loc)
{
	_mouseDown = false;
	SetEventMask(0);

	//if it was just a click (and not a drag)
	if (loc == _mouseDownPosition)
	{
		//if it isn't already snapped, snap it
		if (!_snapped)
		{
			Close();
		}
		else
		{ //unsnap it by going to previous position
			Open();
		}
		
	}
	else Invalidate(); //redraw the splitter in 'up' position.
}

void ZSplitPane::Draw(BRect update)
{
	if (Window()->Lock())
	{
		//just call the DrawSplitter routine
		DrawSplitter(_splitterRect, _orientation, _mouseDown);
		Window()->Unlock();
	}
}

void ZSplitPane::DrawSplitter(BRect splitter, long orientation, bool mouseDown)
{	
	rgb_color bg = ViewColor();
	rgb_color lighter = tint_color(bg, 0.25F);
	rgb_color darker = tint_color(bg, B_DARKEN_3_TINT);

	SetHighColor(bg);
	FillRect(splitter);
	
	SetPenSize(0);

	BRect nubbin;	
	
	if (orientation == HORIZONTAL)
	{
		nubbin.top = splitter.Height() / 2 - 10;
		nubbin.bottom = nubbin.top + 20;	
		nubbin.left = splitter.left;
		nubbin.right = nubbin.left + 2;

		long drawStyle = mouseDown ? SUNKEN : RAISED;
		if (_snapped) drawStyle = FLUSH;
		
		_draw3DRect(this, nubbin, ViewColor(), lighter, darker, drawStyle);
	
		nubbin.OffsetBy(5, 0);
		_draw3DRect(this, nubbin, ViewColor(), lighter, darker, drawStyle);
	}
	else
	{
		nubbin.top = splitter.top;
		nubbin.bottom = nubbin.top + 2;
		nubbin.left = splitter.Width() / 2 - 10;
		nubbin.right = nubbin.left + 20;

		long drawStyle = mouseDown ? SUNKEN : RAISED;
		if (_snapped) drawStyle = FLUSH;
		
		_draw3DRect(this, nubbin, ViewColor(), lighter, darker, drawStyle);
	
		nubbin.OffsetBy(0, 5);
		_draw3DRect(this, nubbin, ViewColor(), lighter, darker, drawStyle);
	}	
}	



void ZSplitPane::SetSplitPosition(float where)
{
	BRect b(Bounds());
	
	if (_orientation == HORIZONTAL)
	{
		if (where < 0) where = b.right + where;
	
		_splitterRect.top = 0;
		_splitterRect.bottom = b.bottom;
		_splitterRect.left = where - (_splitterWidth /  2);
		_splitterRect.right = where + (_splitterWidth /  2);
		
		//should splitter be snapped?
		if (_splitterRect.right > b.right - _snapDistance)
		{
			_splitterRect.right = b.right;
			_splitterRect.left = _splitterRect.right - _splitterWidth;

			if (!_snapped)
			{
				_snapped = true;			
				_snappedAt = SNAP_SECOND;
			}
		}
		else if (_splitterRect.left < _snapDistance)
		{
			_splitterRect.left = 0;
			_splitterRect.right = _splitterWidth;

			if (!_snapped)
			{
				_snapped = true;
				_snappedAt = SNAP_FIRST;
			}
		}
		else _snapped = false;
	}
	else
	{
		if (where < 0) where = b.bottom + where;
		
		_splitterRect.top = where - (_splitterWidth /  2);
		_splitterRect.bottom = where + (_splitterWidth /  2);
		_splitterRect.left = 0;
		_splitterRect.right = b.right;
		
		//should splitter be snapped?
		if (_splitterRect.bottom > b.bottom - _snapDistance)
		{
			_splitterRect.bottom = b.bottom;
			_splitterRect.top = _splitterRect.bottom - _splitterWidth;

			if (!_snapped)
			{
				_snapped = true;			
				_snappedAt = SNAP_SECOND;
			}
		}
		else if (_splitterRect.top < _snapDistance)
		{
			_splitterRect.top = 0;
			_splitterRect.bottom = _splitterWidth;

			if (!_snapped)
			{
				_snapped = true;
				_snappedAt = SNAP_FIRST;
			}
		}
		else _snapped = false;
	}
	
	if (_snapped) _stopAnimation();
			
	//layout children with the changes we just made
	_doLayout();
}

float ZSplitPane::SplitPosition()
{
	if (_orientation == HORIZONTAL)
		return (_splitterRect.right + _splitterRect.left) / 2;
	else
		return (_splitterRect.bottom + _splitterRect.top) / 2;
}

void ZSplitPane::_doLayout()
{

	BRect bounds = Bounds();
	BRect firstChildFrame(bounds), secondChildFrame(bounds);

	if (_snapped)
	{
		if (_orientation == HORIZONTAL)
		{
			if (_snappedAt == SNAP_FIRST)
			{
				_splitterRect = bounds;
				_splitterRect.right = _splitterWidth;
			}
			else
			{
				_splitterRect = bounds;
				_splitterRect.left = _splitterRect.right - _splitterWidth;
			}
		}
		else
		{
			if (_snappedAt == SNAP_FIRST)
			{
				_splitterRect = bounds;
				_splitterRect.bottom = _splitterWidth;
			}
			else
			{
				_splitterRect = bounds;
				_splitterRect.top = _splitterRect.bottom - _splitterWidth;
			}
		}
	}

	//now actually layout the children's rectangles
	if (_orientation == HORIZONTAL)
	{
		firstChildFrame.right = _splitterRect.left - 1;
		secondChildFrame.left = _splitterRect.right + 1;		
	}
	else
	{
		firstChildFrame.bottom = _splitterRect.top - 1;
		secondChildFrame.top = _splitterRect.bottom + 1;
	}

	
	//now position the first and second child accordingly
	if (_firstChild && (firstChildFrame != _firstChild->Frame()))
	{
		_firstChild->MoveTo(firstChildFrame.left, firstChildFrame.top);
		_firstChild->ResizeTo(firstChildFrame.Width(), firstChildFrame.Height());
	}

	if (_secondChild && (secondChildFrame != _secondChild->Frame()))
	{
		_secondChild->MoveTo(secondChildFrame.left, secondChildFrame.top);
		_secondChild->ResizeTo(secondChildFrame.Width(), secondChildFrame.Height());
	}	
	
	Invalidate();
}

void ZSplitPane::SetFollowingMode(long follow)
{
	if ((follow == FOLLOW_FIRST) || (follow == FOLLOW_SECOND))
		_follow = follow;	

	_doLayout();
}

long ZSplitPane::FollowingMode()
{
	return _follow;
}

void ZSplitPane::SetOrientation(long orient)
{
	if (orient == HORIZONTAL)
	{
		_orientation = HORIZONTAL;

		//'rotate' splitter into new position
		BRect newSplitter;
		newSplitter.top = 0;
		newSplitter.bottom = Bounds().bottom;
		newSplitter.left = _splitterRect.top;
		newSplitter.right = newSplitter.left + _splitterWidth;

		_splitterRect = newSplitter;		
	}
	else if (orient == VERTICAL)
	{
		_orientation = VERTICAL;

		//'rotate' splitter into new position
		BRect newSplitter;
		newSplitter.top = _splitterRect.left;
		newSplitter.bottom = newSplitter.top + _splitterWidth;
		newSplitter.left = 0;
		newSplitter.right = Bounds().right;

		_splitterRect = newSplitter;
	}
	
	_doLayout();
}

long ZSplitPane::Orientation()
{
	return _orientation;
}

void ZSplitPane::_stopAnimation()
{
	if (_messageRunner){
		delete _messageRunner;
		_messageRunner = NULL;
	}
}

void ZSplitPane::MessageReceived(BMessage *msg)
{
	switch(msg->what)
	{
		case ZSPLIT_PANE_STEP_CLOSE:
		{
		
			if (_orientation == HORIZONTAL)
			{
				if (_follow == FOLLOW_FIRST)
				{
					float pos = SplitPosition();
					
					if (pos > 2)
						SetSplitPosition(pos / 2);
					else
					{
						_stopAnimation();
						SetSplitPosition(0);
					}
				}
				else
				{
					float right = Bounds().right;
					
					float pos = SplitPosition();
					float dx = right - pos;
					
					if (pos < right - 2)
						SetSplitPosition(pos + dx / 2);
					else
					{
						_stopAnimation();
						SetSplitPosition(right);
					}
				}
			}
			else
			{
				if (_follow == FOLLOW_FIRST)
				{
					float pos = SplitPosition();
					
					if (pos > 2)
						SetSplitPosition(pos / 2);
					else
					{
						_stopAnimation();
						SetSplitPosition(0);
					}
				}
				else
				{
					float bottom = Bounds().bottom;
					
					float pos = SplitPosition();
					float dy = bottom - pos;
					
					if (pos < bottom - 2)
						SetSplitPosition(pos + dy / 2);
					else
					{
						_stopAnimation();
						SetSplitPosition(bottom);
					}
				}
			}
					
			break;
		}
		
		default: BView::MessageReceived(msg);
	}
}

void ZSplitPane::Open()
{
	_snapped = false;
	_splitterRect = _preSnapSplitterRect;
	_doLayout();		
}

void ZSplitPane::Close()
{
	if (!_snapped)
	{
		if (_messageRunner) delete _messageRunner;
		_messageRunner = new BMessageRunner(this, new BMessage(ZSPLIT_PANE_STEP_CLOSE), 100);
	}
}